from datetime import datetime, timedelta

type_dict = {
    'str': type(''),
    'int': type(0),
    'dict': type({}),
    'list': type([])
}
re_type_dict = {
    type(''): 'str',
    type(0): 'int',
    type({}): 'dict',
    type([]): 'list'
}

def validate_type(*args, **kwargs):
    '''
    Validating Data Type

    Description: 
        Validate type of one / more variable with one / more data type class

    Usage:
        0. Sample data
            variable = 'test data'
            variable
            # test data

            variable2 = 4
            variable2
            # 4

        1.  Check only one variable with certain data type:
            data_type(variable, str)
            # None

        2.  Check only one variable with multiple possible data type:
            data_type(variable, str, int)
            # None

        3.  Check two / more variables with one / multiple possible data type:
            data_type([
                (variable, str, int),
                (variable2, int)    
            ])
            # None

        4.  When the data has different accepted type(s):
            data_type(variable, int, list)
            # "Data has invalid accepted type! (Index: 0, Input: str, Accepted: int, list)"

    '''
    checks = []
    if len(args) == 0:
        pass
    elif len(args) == 1:
        checks = args[0]
    elif len(args) >= 2:
        check = []
        for arg in args:
            if arg != args[0]:
                if type(arg) != type(type('')):
                    raise ReferenceError("Doesn't accept another argument after a 'dict' argument!")
            check.append(arg)
        checks.append(tuple(check))
    for check in checks:
        index = checks.index(check)
        result = []
        check = list(check)
        type_input = type(check[0])
        check.pop(0)
        for validation in check:
            if type(validation) == type(''):
                result.append(type_input == type_dict[validation])
            else:
                result.append(type_input == validation)
        if True not in result:
            error_input_string = re_type_dict[type_input]
            error_args_string = ", ".join(repr(re_type_dict[v]) for v in check).replace("'","").replace('"','')
            if kwargs.get('name') == None:
                raise TypeError("Data has invalid accepted type! (Index: %s, Input: %s, Accepted: %s)"%(index, error_input_string, error_args_string))
            else:
                raise TypeError("'%s' data has invalid accepted type! (Index: %s, Input: %s, Accepted: %s)"%(kwargs.get('name'), index, error_input_string, error_args_string))

def utcnow():
    return datetime.utcnow()

def now(**kwargs):
    if len(kwargs) == 0:
        return datetime.now()
    elif len(kwargs) == 1:
        return utcnow() + timedelta(kwargs)
    else:
        pass
    


from sqlite3 import connect
import psycopg2 as pc
import redshift_connector as rc
try:
    from vetl import tool
except:
    import sys
    sys.path.insert(1, '/Users/asyraf/Documents/_Asyraf/# PRIVATE/Workspace/git/asyrafadianto_vetl')
    from vetl import tool

data = {
    "credential":{},
    "session":{}
}
default = {
    "port": {
        "5432": "postgres",
        "5439": "redshift"
    }
}
model = {
    "parameter": {
        "host": str,
        "port": int,
        "dbname": str,
        "username": str,
        "password": str,
        "engine": str
    }
}
message = {
    "error": {
        "credential":{
            "user_args_and_kwargs_together": "Cannot run when both args and kwargs are provided!",
            "user_args_too_much": "Too much args provided!",
            "user_args_mistype": "Invalid data type of provided args!",
            "internal_update_less_input": "Required 'host', 'port', 'dbname', 'username' and 'password' but only got %s!"
        }
    }
}

def start(connection_name):
    pass

def credential(connection_name, *args, **kwargs):
    '''
    Credential CRUD

    Create (Done)
        Provide at least the 'connection_name'
            Example:
                credential('test')
                ~#  <timestamp> Log: Create database credential with connection name 'test' success!

    Retrieve (Done)
        Provide at least the 'connection_name'
            Example:
                print(credential('test'))
                ~#  ...
                #   {}

    Update
        1. Provide a 'connection_name' and multiple 'kwargs' value
            Example:
                credential('test', host='1', port=0, ...)
                ~#  <timestamp> Log: Update database credential (host: 'host', port: 'port', ...) for the connection name 'test' success!
        2. Provide a 'connection_name' and a single 'args' value with 'dict' type
            Example:
                credential('test', {'host':'1', 'port':0, ...})
                ~#  <timestamp> Log: Update database credential (host: 'host', port: 'port', ...) for the connection name 'test' success!

    Delete
        Provide a 'connection_name' and a 'delete' args
            Example:
                credential('test', delete)
                ~#  <timestamp> Log: Delete database credential with connection name 'test' success!

    
    '''
    if len(args) == 0:
        if (len(kwargs) == 0): # Provide only the connection_name
            return _credential_retrieve_(connection_name)
        elif len(kwargs) != 0: # Provide a connection_name and multiple kwargs -> ..., host='', port=0, ...
            if _credential_retrieve_(connection_name) == None:
                _credential_create_(connection_name)
            ckwargs = {}
            for key, value in kwargs.items():
                ckwargs[key] = value
            _credential_update_(connection_name, ckwargs)
    elif len(args) == 1: 
        if len(kwargs) == 0: # Provide connection_name and only one structured args -> ..., {'host':'', 'port':0, ...}
            if type(args[0]) == type({}) :
                if _credential_retrieve_(connection_name) == None:
                    _credential_create_(connection_name)
                _credential_update_(connection_name, args[0])
            elif args[0] == 'delete':
                _credential_delete_(connection_name)
            else:
                raise ValueError(message['error']['credential']['user_args_mistype'])
        elif len(kwargs) != 0:
            raise ValueError(message['error']['credential']['user_args_and_kwargs_together'])
    else:
        raise ValueError(message['error']['credential']['user_args_too_much'])

def _credential_create_(connection_name):
    global data
    data['credential'][connection_name] = {}
    print("<%s> Log: Create database credential with connection name '%s' success!"%(tool.utcnow(), connection_name))

def _credential_retrieve_(connection_name):
    global data
    return data['credential'].get(connection_name)

def _credential_update_(connection_name, *args):
    global data
    required = ['host', 'port', 'dbname', 'username', 'password']
    for key, value in args[0].items():
        if key in required:
            tool.validate_type(value, model['parameter'][key])
            data['credential'][connection_name][key] = value
            if key == 'port':
                if 'engine' in args[0].keys():
                    data['credential'][connection_name]['engine'] = args[0]['engine']
                else:
                    data['credential'][connection_name]['engine'] = default['port'][str(value)]
    print(
        "<%s> Log: Update database credential (host: %s, port: %s, dbname: %s, username: %s, engine: %s) for the connection name '%s' success!"%(
            tool.utcnow()
            ,data['credential'][connection_name].get('host')
            ,data['credential'][connection_name].get('port')
            ,data['credential'][connection_name].get('dbname')
            ,data['credential'][connection_name].get('username')
            ,data['credential'][connection_name].get('engine')
            ,connection_name
        )
    )

def _credential_delete_(connection_name):
    global data
    data['credential'].pop(connection_name)
    print("<%s> Log: Delete database credential with connection name '%s' success!"%(tool.utcnow(), connection_name))

def test():
    try:
        credential(
            'rds-stag'
            ,{
                'host': 'renos-v2-staging-rds.cssnfgwcjwmm.ap-southeast-1.rds.amazonaws.com'
                ,'port': 5432
                ,'dbname': 'renosdb'
                ,'username': 'aws_glue_staging'
                ,'password': '/>jF>YS-22;6x{vHNV.nnq)"_p-h}N7&'
            }
        )
        credential('rds-stag', host='a')
        print(credential('rds-stag'))
    except Exception as e:
        print('Error: %s'%(e))
test()


from setuptools import setup, find_packages
  
with open("README.md", "r") as fh:
    description = fh.read()
  
setup(
    name="pypetl",
    version="0.0.2",
    author="Asyraf N. Adianto",
    author_email="asyraf.adianto@renos.id",
    package_dir={'': '.'},
    packages=find_packages('.'),
    description="Python ETL Tools",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://github.com/asyraf-adianto/pypetl",
    license='MIT',
    python_requires='>=3.0',
    install_requires=[
        'pandas>=1.0.3',
        'petl>=1.7.0',
        'redshift-connector>=2.0.908',
        'psycopg2-binary>=2.9.3',
        'paramiko>=2.11.0',
        'sshtunnel>=0.4.0'
    ]
)